@@ -3,7 +3,7 @@
  * @(#)lex.c
  *
  * Copyright 1997-1999, Wes Cherry   (mailto:wesc@technosis.com)
- *           2000-2001, Aaron Ardiri (mailto:aaron@ardiri.com)
+ *           2000-2003, Aaron Ardiri (mailto:aaron@ardiri.com)
  * All rights reserved.
  * 
  * This program is free software; you can redistribute it and/or modify
@@ -120,8 +120,6 @@ static BOOL
 FParseHex(LEX * plex,
           int ch)
 {
-  LEX lex;
-
   ch = tolower(ch);
   if ((ch == '0') && ((*(pchLex) == 'x') || (*(pchLex) == 'X')))
   {
@@ -130,21 +128,41 @@ FParseHex(LEX * plex,
   }
   if ((ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'f'))
   {
-    lex.lt = ltConst;
-    lex.val = 0;
+	plex->lt = ltConst;
+	plex->val = 0;
     while ((ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'f'))
     {
-      lex.val *= 16;
+      plex->val *= 16;
 
       if ((ch >= '0' && ch <= '9'))
-        lex.val += ch - '0';
+        plex->val += ch - '0';
       else
-        lex.val += ch - 'a' + 10;
+        plex->val += ch - 'a' + 10;
       ch = *pchLex++;
       ch = tolower(ch);
     }
+
+   plex->size = lsUnspecified;
+   if (ch == '.')
+   {
+    if (*pchLex == 'b' || *pchLex == 'B')
+    {
+     plex->size = lsByte;
+     pchLex += 2;
+    }
+    else if (*pchLex == 'w' || *pchLex == 'W')
+    {
+     plex->size = lsWord;
+     pchLex += 2;
+    }
+    else if (*pchLex == 'l' || *pchLex == 'L')
+    {
+     plex->size = lsLong;
+     pchLex += 2;
+    }
+   }
+
     AllowLUAtEndOfConstant(ch);
-    *plex = lex;
     return fTrue;
   }
   return fFalse;
@@ -177,6 +195,26 @@ FParseConst(LEX * plex,
       *pchStore++ = (char)ch;
       ch = *pchLex++;
     }
+
+    plex->size = lsUnspecified;
+    if (ch == '.')
+    {
+      if (*pchLex == 'b' || *pchLex == 'B')
+      {
+       plex->size = lsByte;
+       pchLex += 2;
+      }
+      else if (*pchLex == 'w' || *pchLex == 'W')
+      {
+       plex->size = lsWord;
+       pchLex += 2;
+      }
+      else if (*pchLex == 'l' || *pchLex == 'L')
+      {
+       plex->size = lsLong;
+       pchLex += 2;
+      }
+    }
   }
   else if (ch == '\'')
   {
@@ -327,7 +365,29 @@ FParseId(LEX * plex,
     while ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z')
            || (ch >= '0' && ch <= '9') || ch == '_');
     lex.szId[cch] = '\000';
+
     *plex = lex;
+
+    plex->size = lsUnspecified;
+    if (ch == '.')
+    {
+      if (*pchLex == 'b' || *pchLex == 'B')
+      {
+       plex->size = lsByte;
+       pchLex += 2;
+      }
+      else if (*pchLex == 'w' || *pchLex == 'W')
+      {
+       plex->size = lsWord;
+       pchLex += 2;
+      }
+      else if (*pchLex == 'l' || *pchLex == 'L')
+      {
+       plex->size = lsLong;
+       pchLex += 2;
+      }
+    }
+
     return fTrue;
   }
   return fFalse;
@@ -369,7 +429,7 @@ FGetLex(LEX * plex,
         BOOL fInComment)
 {
   int ch;
-  char szT[2];
+  char szT[4];
   LEX lex;
   char *pchStore;
 
@@ -399,6 +459,7 @@ FGetLex(LEX * plex,
         /*
          * TODO logical operators 
          */
+        SLT(0xA0, ltConst)                       /* RMa add to support TRAP special case */
       SLT('.', ltPoint) SLT('+', ltPlus) SLT('-', ltMinus) SLT('*', ltMult) SLT('%', ltMod) SLT('(', ltLParen) SLT(')', ltRParen) SLT('[', ltLBracket) SLT(']', ltRBracket) SLT('{', ltLBrace) SLT('}', ltRBrace) SLT(',', ltComma) SLT('?', ltQuestion) SLT(':', ltColon) SLT('^', ltCaret) SLT('\\', ltBSlash) SLT('#', ltPound) SLT('@', ltAt) SLT(';', ltSemi) SLT('|', ltPipe) case '/':
         if (*pchLex == '/')
         {
@@ -565,8 +626,10 @@ FGetLex(LEX * plex,
         }
         else
         {
-          szT[0] = (char)ch;
-          szT[1] = '\000';
+          szT[0] = '\'';
+          szT[1] = (char)ch;
+          szT[2] = '\'';
+          szT[3] = '\000';
           ParseError("Unknown character: ", szT);
         }
         pchLex--;
